#Author: Marc Neu
#Date: 01.12.2021

import numpy
from b2tsf.belle_CDC_layer import CDCLayer

class CDCFrame:
    """
    The class CDCFrame represents a whole detector data frame of the Belle-II
    Detector. Automatically instantiates all necessary layers.
    ---------------------------------------------------------------------------
    Attributes:
    ---------------------------------------------------------------------------
    _layer:    Internal data structure for layers

    event_id:   Contains event id. This attribute is
                used to create a data time series when simulating dynamic
                trigger characteristics.
    ---------------------------------------------------------------------------
    Methods:
    ---------------------------------------------------------------------------
    get_layer:      Get single layer by index

    get_layers:     Get all layers as list

    get_hit_list:   Returns list of all hits in eihter native format, e.g.
                    (layer_id,segment_id, row_id, column_id) or in export format
                    (layer_id,row_id,theta_id)

    get_number_of_hits:
                    Returns number of hits as length of get_hit_list()

    clear:          Sets internal frame buffer to zero

    copy:           Deep copy of frame supplied as parameter

    and:            logical and between frame and frame supplied as parameter

    populate:       Takes a list of hit addresses. This
                    addressing scheme fits the data supplied by MPG Group.
    """


    def __init__(self,event_id=0):
        self._layer = [ CDCLayer(i) for i in range(9)]
        self.event_id = event_id


    def __calc_index(self,extern_layer_id,extern_azimuth_id):
        #Init values
        layer_id = 0
        seg_id = 0
        row_id = 0
        col_id = 0
        temp_id = 0

        #Find superlayer
        for i in range(9):
            if(extern_layer_id < temp_id + self.get_layer(i).n_rows):
                layer_id = i
                break;
            else:
                temp_id = temp_id + self.get_layer(i).n_rows

        extern_layer_id = extern_layer_id - temp_id
        #Find Row
        row_id = extern_layer_id

        temp_id = 0
        #Read values from superlayer
        n_rows = self.get_layer(layer_id).n_rows
        n_cols = self.get_layer(layer_id).n_cols
        n_segs = self.get_layer(layer_id).n_segs

        #Find Segment
        for i in range(n_segs):
            if(extern_azimuth_id < temp_id + n_cols):
                seg_id = i
                break;
            else:
                temp_id = temp_id + n_cols

        extern_azimuth_id = extern_azimuth_id - temp_id
        #Find Column
        col_id = extern_azimuth_id

        return (int(layer_id), int(seg_id), int(row_id), int(col_id))


    def get_layers(self):
        return self._layer


    def get_layer(self,layer_id):
        return self._layer[layer_id]


    def get_hit_list(self,format='native',detect_hit=True,skip_inner_layer=False):
        if(format == 'native'):
            layer_id_list = []
            seg_id_list = []
            row_id_list = []
            col_id_list = []
            for i in range(skip_inner_layer,9):
                layer = self._layer[i]
                (seg_id,
                 row_id,
                 col_id) = numpy.nonzero(layer.get_segments() == detect_hit)
                l_hits = numpy.count_nonzero(layer.get_segments() == detect_hit)

                layer_id_list.extend(l_hits * [layer.layer_id])
                seg_id_list.extend(list(seg_id.astype(int)))
                row_id_list.extend(list(row_id.astype(int)))
                col_id_list.extend(list(col_id.astype(int)))

            result = (layer_id_list,seg_id_list,row_id_list,col_id_list)

        elif(format == 'r_theta'):
            layer_id_list = []
            row_id_list = []
            theta_id_list = []
            for i in range(skip_inner_layer,9):
                layer = self._layer[i]
                (seg_id,
                 row_id,
                 col_id) = numpy.nonzero(layer.get_segments() == detect_hit)
                theta_id = seg_id*layer.n_cols+col_id
                l_hits = numpy.count_nonzero(layer.get_segments() == detect_hit)

                layer_id_list.extend(l_hits * [layer.layer_id])
                row_id_list.extend(list(row_id.astype(int)))
                theta_id_list.extend(list(theta_id.astype(int)))

            result = (layer_id_list,row_id_list,theta_id_list)

        else:
            pass

        return result


    def get_number_of_hits(self):
        n_hits = 0
        for layer in self._layer:
            l_hits = numpy.count_nonzero(layer.get_segments())
            n_hits = n_hits + l_hits

        return n_hits

    def fill(self):
        for layer in self._layer:
            layer.fill()
        return self


    def clear(self):
        for layer in self._layer:
            layer.clear()
        return self


    def copy(self,other):
        for (layer,other_layer) in zip(self._layer,other.get_layers()):
            layer = layer.copy(other_layer)
        return self


    def logical_and(self,other):
        for (layer,other_layer) in zip(self._layer,other.get_layers()):
            layer =  layer.logical_and(other_layer)
        return self

    def logical_or(self,other):
        for (layer,other_layer) in zip(self._layer,other.get_layers()):
            layer =  layer.logical_or(other_layer)
        return self


    def populate(self,layer_id_list,azimuth_id_list):
        for extern_layer_id,extern_azimuth_id in zip(layer_id_list,azimuth_id_list):
            #Read data position
            (layer_id, seg_id, row_id, col_id) = self.__calc_index(extern_layer_id,extern_azimuth_id)
            #Write data
            self.get_layer(layer_id).get_segment(seg_id)[row_id,col_id] = 1



import numpy
from bitarray import bitarray
from bitarray import frozenbitarray
import warnings
from b2tsf.belle_CDC_frame import CDCFrame

class TrackSegmentFinderLUT5:
    """
    Track Segement Finder implenmenting a 5 bit look up table solution.

    ---------------------------------------------------------------------------
    Methods:
    ---------------------------------------------------------------------------
    detect:     Detects hits in incoming frame <frame_in> and saves result in
                internal data strucutre. Two possible detetction modes are
                available. 'hits' returns real hardware hits whereas 'candidates'
                returns all hit candidates of a hit bin.

    get_frame:  Retruns internal data frame as reference.

    clear:      Clears internal data structure.

    """
    def __init__(self):
        self._frame_out = CDCFrame(0)
        self._binsize = 5
        self._binnumber = 16
        self._binlayer = 2
        self.mode = 'hits'
        self._patterns = self._init_patterns([
            '11111','10011','00111','11001','01110','11100'])

        self._init = False
        self._frame_in = None
        self._frame_in_delayed = None
        self._cooldown_array = None
        self._cooldown = 64
        self._n_tsout = 24

    def _init_patterns(self,patterns):
        keys = []
        values = []
        for i in range(2**self._binsize):
            key_pattern = f'{i:05b}'
            keys.append(frozenbitarray(key_pattern))
            values.append(key_pattern in patterns)
        return dict(zip(keys,values))


    def _detect_hit(self,hitmap_zip,seg_hits):
        """
        Implementation of per bin detetion algorithm, e.g. checking against
        look up table in form of a hash map. hitmap_zip contains both bins
        for upper and lower hitmap.
        This function requires a predefined array for performance reasons. This
        way the creation of many small arrays can be avoided by slicing.
        """
        hitmap_lower, hitmap_upper = hitmap_zip
        for i in range(self._binnumber):
            seg_hits[1,i] = self._patterns[hitmap_upper[i]]
            seg_hits[0,i] = self._patterns[hitmap_lower[i]]
        return seg_hits


    def _populate_bin(self,seg_pre, seg_act, seg_suc, edge_hit=True):
        """
        Populate bins for detection algorithm. Each segment has got 32 bins as
        specified in the alogrithm description in thesis. Furthermore edge_hits
        of sectors can be toggeled on/off.
        This function returns bins of a layer segment.
        """
        hit_bin_lower = [self._binsize * bitarray() for i in range(self._binnumber)]
        hit_bin_upper = [self._binsize * bitarray() for i in range(self._binnumber)]

        if(edge_hit):
            # Circulation counterclockwise - successor left, predecessor right
            hit_bin_lower[0] = frozenbitarray([
                seg_act[1,0],seg_pre[1,15],
                seg_act[0,1],seg_act[0,0],seg_pre[0,15]])
            hit_bin_lower[15] = frozenbitarray([
                seg_act[1,15],seg_act[1,14],
                seg_suc[0,0],seg_act[0,15],seg_act[0,14]])
            hit_bin_upper[0] = frozenbitarray([
                seg_act[4,1],seg_act[4,0],seg_pre[4,15],
                seg_act[3,0],seg_pre[3,15]])
            hit_bin_upper[15] = frozenbitarray([
                seg_suc[4,0],seg_act[4,15],seg_act[4,14],
                seg_act[3,15],seg_act[3,14]])
        else:
            hit_bin_lower[0] = frozenbitarray([
                seg_act[1,0],0,
                seg_act[0,1],seg_act[0,0],0])
            hit_bin_lower[15] = frozenbitarray([
                seg_act[1,15],seg_act[1,14],
                0,seg_act[0,15],seg_act[0,14]])
            hit_bin_upper[0] = frozenbitarray([
                seg_act[4,1],seg_act[4,0],0,
                seg_act[3,0],0])
            hit_bin_upper[15] = frozenbitarray([
                0,seg_act[4,15],seg_act[4,14],
                seg_act[3,15],seg_act[3,14]])

        for i in range(0+1,self._binnumber-1):
            hit_bin_lower[i] = frozenbitarray([
                seg_act[1,i],seg_act[1,i-1],
                seg_act[0,i+1],seg_act[0,i],seg_act[0,i-1]])

        for i in range(0+1,self._binnumber-1):
            hit_bin_upper[i] = frozenbitarray([
                seg_act[4,i+1],seg_act[4,i],seg_act[4,i-1],
                seg_act[3,i],seg_act[3,i-1]])

        return (hit_bin_lower,hit_bin_upper)


    def _check_layer_hit(self,layer_in, edge_hit=True):
        """
        Checks frame layer <layer_in> for hits. Using edge_hit True/False segemnt edge mode can be
        choosen. For production use edge_hit=True as non edge_hit is only used for hardware module
        testing.
        """
        n_segs = layer_in.n_segs
        hits = numpy.zeros(shape=(n_segs,self._binlayer,self._binnumber),dtype=int)
        # Segment 0
        seg_pre = layer_in.get_segment(n_segs-1)
        seg_act = layer_in.get_segment(0)
        seg_suc = layer_in.get_segment(1)
        hit_bin = self._populate_bin(seg_pre, seg_act, seg_suc,edge_hit)
        hits[0,:,:] = self._detect_hit(hit_bin, hits[0,:,:])
        # Inner segments
        for i in range (1,n_segs-1):
            seg_pre = layer_in.get_segment(i-1)
            seg_act = layer_in.get_segment(i)
            seg_suc = layer_in.get_segment(i+1)
            hit_bin = self._populate_bin(seg_pre, seg_act, seg_suc,edge_hit)
            hits[i,:,:] = self._detect_hit(hit_bin, hits[i,:,:])
        # Segment n_seg
        seg_pre = layer_in.get_segment(n_segs-2)
        seg_act = layer_in.get_segment(n_segs-1)
        seg_suc = layer_in.get_segment(0)
        hit_bin = self._populate_bin(seg_pre, seg_act, seg_suc,edge_hit)
        hits[-1,:,:] = self._detect_hit(hit_bin, hits[-1,:,:])

        return hits


    def _detect_hits(self,frame_in,edge_hit=True):
        """
        Detects hits of all layers in <frame_in> and saves output in tsf frame.
        Hardcoded to layer 1...8, because hardware implementation for layer 0 is
        non existent. This feature could be implemented later.

        Step 1: Calculate hits for each layer. Returns 2D Array of hits
                with indices <segment_id> whose range depends on the layer
                and 0...31 for each checked hit bin inside the dection algorithm.
                E.g. Bin 5 in Segment 3 was hit => hits[2,4] == True

        Step 2: Find all hits and fill output frame accordingly.
        """
        self._frame_in = frame_in
        for layer_id in range(1,9):
            layer_in = frame_in.get_layer(layer_id)
            layer_segments = self._frame_out.get_layer(layer_id).get_tsf_segments()
            hits = self._check_layer_hit(layer_in,edge_hit)
            (seg_id_list,row_id_list,col_id_list) = numpy.nonzero(hits)
            for seg_id,row_id,col_id in zip(seg_id_list,row_id_list,col_id_list):
                layer_segments[seg_id,0+row_id*4,col_id] = 1
        return self._frame_out


    def _detect_candidates(self,frame_in,edge_hit=True):
        """
        Detects hits of all layers in <frame_in> and saves output in tsf frame.
        Hardcoded to layer 1...8, because hardware implementation for layer 0
        is non existent. This feature could be implemented later.

        Step 1: Calculate hits for each layer. Returns 2D Array of hits
                with indices <segment_id> whose range depends on the layer
                and 0...31 for each checked hit bin inside the dection
                algorithm, e.g. Bin 5 in Segment 3 was hit => hits[2,4] == True

        Step 2: Find all hits and fill output frame accordingly.

        Identical to _detect_hits() except that it populates all hit bin
        candidates to the internal output data frame.
        """
        self._frame_in = frame_in
        for layer_id in range(1,9):
            layer_in = frame_in.get_layer(layer_id)
            layer_segments = self._frame_out.get_layer(layer_id).get_tsf_segments()
            #Calculate hits of layer
            hits = self._check_layer_hit(layer_in,edge_hit)
            (seg_id_list,up_lo_id,col_id_list) = numpy.nonzero(hits)
            for seg_id,up_lo_id,col_id in zip(seg_id_list,up_lo_id,col_id_list):
                if(up_lo_id == 0):
                    if(col_id == 0):
                        layer_segments[seg_id-1,0,15] = 1
                        layer_segments[seg_id,0,0] = 1
                        layer_segments[seg_id,0,1] = 1
                        layer_segments[seg_id-1,1,15] = 1
                        layer_segments[seg_id,1,0] = 1
                    elif(col_id == 15):
                        layer_segments[seg_id,0,14] = 1
                        layer_segments[seg_id,0,15] = 1
                        layer_segments[(seg_id+1)%layer_in.n_segs,0,0] = 1
                        layer_segments[seg_id,1,14] = 1
                        layer_segments[seg_id,1,15] = 1
                    else:
                        layer_segments[seg_id,0,col_id-1] = 1
                        layer_segments[seg_id,0,col_id] = 1
                        layer_segments[seg_id,0,col_id+1] = 1
                        layer_segments[seg_id,1,col_id-1] = 1
                        layer_segments[seg_id,1,col_id] = 1
                else:
                    if(col_id == 0):
                        layer_segments[seg_id-1,4,15] = 1
                        layer_segments[seg_id,4,0] = 1
                        layer_segments[seg_id,4,1] = 1
                        layer_segments[seg_id-1,3,15] = 1
                        layer_segments[seg_id,3,0] = 1
                    elif(col_id == 15):
                        layer_segments[seg_id,4,14] = 1
                        layer_segments[seg_id,4,15] = 1
                        layer_segments[(seg_id+1)%layer_in.n_segs,4,0] = 1
                        layer_segments[seg_id,3,14] = 1
                        layer_segments[seg_id,3,15] = 1
                    else:
                        layer_segments[seg_id,4,col_id-1] = 1
                        layer_segments[seg_id,4,col_id] = 1
                        layer_segments[seg_id,4,col_id+1] = 1
                        layer_segments[seg_id,3,col_id-1] = 1
                        layer_segments[seg_id,3,col_id] = 1
        return self._frame_out

    def _detect_hw_model(self,frame_in,edge_hit=True):
        """
        Detects hit in a sequence, if Track Segment Finder is called multiple
        times. Returns frame with hit data.
        """
        if(not self._init):
            self._frame_in_delayed = CDCFrame(0)
            #24 is the maximum number of segments, e.g. n_segs in layer 9.
            self._cooldown_array = numpy.zeros(shape=(9,
                                                      24,
                                                      self._binlayer,
                                                      self._binnumber),
                                               dtype=int)
            self._init = True
        else:
            self._frame_in_delayed = self._frame_in_delayed.copy(self._frame_in)
            self._cooldown_array = numpy.where(self._cooldown_array == 0,
                                                self._cooldown_array,
                                                self._cooldown_array-1)
        self._frame_in = frame_in
        frame_work = self._frame_in_delayed.logical_or(frame_in)

        self._frame_out.clear()

        for layer_id in range(1,9):
            layer_in = frame_work.get_layer(layer_id)
            layer_segments = self._frame_out.get_layer(layer_id).get_tsf_segments()
            #Calculate hits of layer
            hits = self._check_layer_hit(layer_in,edge_hit)

            (seg_id_list,row_id_list,col_id_list) = numpy.nonzero(hits)
            for seg_id,up_lo_id,col_id in zip(seg_id_list,row_id_list,col_id_list):
                if(self._cooldown_array[layer_id,seg_id,up_lo_id,col_id] == 0):
                    self._cooldown_array[layer_id,seg_id,up_lo_id,col_id] = self._cooldown
                    layer_segments[seg_id,0+up_lo_id*4,col_id] = 1
        return self._frame_out


    def detect(self,frame_in,mode='hits',edge_hit=True):
        """
        Interface function for detection algorithm.
        """
        self.mode = mode
        if(mode == 'hits'):
            return self._detect_hits(frame_in,edge_hit)
        elif(mode == 'candidates'):
            return self._detect_candidates(frame_in,edge_hit)
        elif(mode == 'hw_model'):
            return self._detect_hw_model(frame_in,edge_hit)
        else:
            raise ValueError(mode)


    def get_segment_dword(self,segment_id):
        """
        Returns list of segment hits as described in thesis.
        """
        dword_layer_list = []
        for work_layer in self._frame_out.get_layers():
            work_seg = work_layer.get_tsf_segment(segment_id)

            dword = numpy.zeros(self._binlayer*self._binnumber,dtype=int)
            dword[:self._binnumber] = work_seg[0,:]
            dword[self._binnumber:self._binnumber*self._binlayer] = work_seg[4,:]

            dword_layer_list.append(dword.tolist())

        return dword_layer_list


    def hit_compaction(self):
        """
        Returns compact list of hits, models stream compaction implementation.
        See master thesis for description of id sorting.
        Works on internal frame which is poulated by callind detect()
        beforehand.
        Returns:
            List of hit list of layers 0 to 9.
        """
        (layer_id_list,
         seg_id_list,
         row_id_list,
         col_id_list) = self._frame_out.get_hit_list(format='native',
                                                      detect_hit=True,
                                                      skip_inner_layer=True)
        hit_list = [[] for i in range(0,9)]

        for (layer_id,seg_id,row_id,col_id) in zip(layer_id_list,seg_id_list,row_id_list,col_id_list):
            if(row_id < 2):
                hardware_id = seg_id*self._binnumber*self._binlayer + col_id + 1
            else:
                hardware_id = seg_id*self._binnumber*self._binlayer + col_id + self._binnumber + 1
            hit_list[layer_id].append(hardware_id)


        #Hardware is only able to process n_tsout number of track segemnts. All other segments
        #are discarded

        hit_list = [hit_list[i][:self._n_tsout] for i in range(0,9)]

        return hit_list


    def get_frame(self):
        return self._frame_out


    def get_n_tsout(self):
        return self._n_tsout


    def clear(self):
        self._frame_out = self._frame_out.clear()
        self._init = False
        return self._frame_out



from datetime import date
import numpy

class TSFProcessorStimuliTestbenchWriter:
    """
    The TSFProcessorStimuliTestbenchWriter creates a testbench stimuli file for
    a vhdl hardware implemenation of a track segment finder. The data block
    sizes is fixed to 256 bit by use of the dword_size attribute.
    After object creation reference a CDCSequence to the TestbenchWriter and
    start the creation of the stimuli file by calling write_layer.
    ----------------------------------------------------------------------------
    Methods:
    ----------------------------------------------------------------------------
    set_sequence
        Accepts a CDCSequence object.
    write_layer
        Writes data of layer from previous defined CDCSequence to a stimuli
        file.
    """

    def __init__(self):
        self.mode = 'tsf_processor_stimuli'
        self.dword_size = 256
        #TODO: Retrieve from Package
        self.version = 0.1
        self.date = '{:%Y-%m-%d}'.format(date.today())
        self.author = 'Marc Neu'
        self._intro_text = "#FileWriter V. {version} generating a {mode} file\n#\n#Author: {author}\n#Date: {date}\n#\n".format(version=self.version,date=self.date,author=self.author,mode=self.mode)
        self._outro_text = "#EOF."

    def set_sequence(self, sequence):
        """
        Configures CDCSequence which is used as data input for the test bench
        stimuli file.
        """
        self._sequence = sequence
        return self


    def _gen_line_layer_full(self,layer):
        """
        Reads layer and converts data points of all segements into one data word.
        """
        stimuli_list = []
        segment_list = layer.get_tsf_segment_list()

        for seg in segment_list:
            dword_256 = numpy.zeros(self.dword_size,dtype=int)
            dword_256[0:80] = numpy.reshape(seg, 80)
            stimuli_list.extend(dword_256.tolist())

        stimuli_list.reverse()
        stimuli_string = ''.join('{:d}'.format(item) for item in stimuli_list)

        return stimuli_string


    def _write_layer_full(self,path,filename,layer_id):
        """
        Writes a all segments of the specified layer to a file under <path>.
        """
        file_path = "{path}{filename}_{id:d}.dat".format(path=path,filename=filename,id=layer_id)

        with open(file_path,"w") as file_ptr:
            file_ptr.write(self._intro_text)

            for frame in self._sequence.get_frames():
                work_layer = frame.get_layer(layer_id)
                data_line = self._gen_line_layer_full(work_layer)
                file_ptr.write(data_line + "\n")

            file_ptr.write(self._outro_text)


    def write_layer(self,path,filename,layer_id,segment_id=0):
        """
        Writes the sequence of one specified layer to a stimuli
        file under <path>.
        """
        self._write_layer_full(path,filename,layer_id)



#Author: Marc Neu
#Date: 01.12.2021

import numpy

class CDCLayer:
    """
    The class CDCLayer represents a superlayer inside the Belle-II Detector.
    The hit data is packed into a numpy array for better performance. Mind
    that the data format is choosen to keep a direct connection to the hardware
    implementation. Therefore segments for each layer are defined.
    ---------------------------------------------------------------------------
    Attributes:
    ---------------------------------------------------------------------------
    layer_id:   Defines physical position inside the detector. Belle-II
                detector currently features 9 superlayers with predefined
                segment numbers.

    n_rows:     Number of rows. Predefined to 6 for inner layer and 8 for outer
                layers.

    n_cols:     16 columns per segment. Predefined by trigger data format

    n_segs:     Number of segemnts. Defined by layer ID.

    n_wires:    Total number of wires inside specified layer

    _segments: Data storage for hit values. 3D numpy arrays contains data for
                the whole layer. Indices are [segment_id,row_id,column_id].
    """

    def __init__(self, layer_id=0):
        self.layer_id = layer_id
        self.n_rows = self.get_nrows(layer_id)
        self.n_cols = 16
        self.n_segs = self.get_nsegs(layer_id)
        self.n_wires = self.n_rows*self.n_cols*self.n_segs
        self._segments = numpy.zeros(shape=(self.n_segs,self.n_rows,self.n_cols),dtype=int)

    def get_nrows(self,layer_id):
        if(layer_id == 0):
            return 8
        elif(layer_id == 1):
            return 6
        elif(layer_id == 2):
            return 6
        elif(layer_id == 3):
            return 6
        elif(layer_id == 4):
            return 6
        elif(layer_id == 5):
            return 6
        elif(layer_id == 6):
            return 6
        elif(layer_id == 7):
            return 6
        elif(layer_id == 8):
            return 6
        else:
            raise InitializationError('CDCLayer ID out of predefined boundary 0 to 8.')

    def get_nsegs(self,layer_id):
        if(layer_id == 0):
            return 10
        elif(layer_id == 1):
            return 10
        elif(layer_id == 2):
            return 12
        elif(layer_id == 3):
            return 14
        elif(layer_id == 4):
            return 16
        elif(layer_id == 5):
            return 18
        elif(layer_id == 6):
            return 20
        elif(layer_id == 7):
            return 22
        elif(layer_id == 8):
            return 24
        else:
            raise InitializationError('CDCLayer ID out of predefined boundary 0 to 8.')


    def get_tsf_segment_list(self):
        """
        get_tsf_segment_list returns a list of all segments in the layer. Each segment consists of a 2D array.
        Only returns the 5 layers which are used for the trigger detection.
        """
        new_list = [self._segments[j,0:5,:] for j in range(self.n_segs)]
        return new_list

    def get_segment_list(self):
        """
        Returns a list of all segemnts in the layer. Each segment consists of a 2D array.
        """
        new_list = [self._segments[j,:,:] for j in range(self.n_segs)]
        return new_list


    def get_tsf_segments(self):
        """
        Native access to data. Only returns 5 layers which are used for the trigger detection
        """
        return self._segments[:,0:5,:]


    def get_segments(self):
        """
        Return reference to interal data structure
        """
        return self._segments


    def get_tsf_segment(self,segment_id):
        """
        Native access to data. Only returns 5 layers which are used for the trigger detection
        """
        return self._segments[segment_id,0:5,:]


    def get_segment(self,segment_id):
        """
        Return reference to interal data structure
        """
        return self._segments[segment_id,:,:]

    def fill(self):
        """
        Fills hit data from layer by setting all wires to 1.
        """
        self._segments.fill(1)
        return self


    def clear(self):
        """
        Clear hit data from layer by filling all values with 0.
        """
        self._segments.fill(0)
        return self


    def copy(self,other):
        """
        Deep copy of <other> layer data (no reference).
        """
        self.layer_id = other.layer_id
        self.n_rows = self.get_nrows(other.layer_id)
        self.n_cols = 16
        self.n_segs = self.get_nsegs(other.layer_id)
        self.n_wires = self.n_rows*self.n_cols*self.n_segs
        self._segments = numpy.copy(other.get_segments())
        return self


    def logical_and(self,other):
        """
        Perform elementwise logical and on layer hit data and save result in this frame.
        """
        numpy.logical_and(self._segments,other.get_segments(),out=self._segments)
        return self


    def logical_or(self, other):
        """
        Perform elementwise logical or on layer hit data and save result in this frame.
        """
        numpy.logical_or(self._segments,other.get_segments(),out=self._segments)
        return self



from b2tsf.belle_CDC_layer import CDCLayer
from b2tsf.belle_CDC_frame import CDCFrame
from b2tsf.belle_CDC_sequence import CDCSequence
from b2tsf.belle_CDC_canvas import CDCCanvas
from b2tsf.track_segment_finder import TrackSegmentFinder
from b2tsf.track_segment_finder_hourglas import TrackSegmentFinderHourglas
from b2tsf.track_segment_finder_LUT5 import TrackSegmentFinderLUT5
from b2tsf.track_segment_finder_LUT9 import TrackSegmentFinderLUT9
from b2tsf.track_segment_finder_LUT12 import TrackSegmentFinderLUT12
from b2tsf.lut_config import LutConfig
from b2tsf.testbench_writer_factory import TestbenchWriterFactory
from b2tsf.merger_processor_ref_testbench_writer import MergerProcessorRefTestbenchWriter
from b2tsf.merger_processor_stimuli_testbench_writer import MergerProcessorStimuliTestbenchWriter
from b2tsf.tsf_processor_ref_testbench_writer import TSFProcessorRefTestbenchWriter
from b2tsf.tsf_processor_stimuli_testbench_writer import TSFProcessorStimuliTestbenchWriter



from b2tsf.track_segment_finder_LUT5 import TrackSegmentFinderLUT5
from b2tsf.track_segment_finder_LUT9 import TrackSegmentFinderLUT9
from b2tsf.track_segment_finder_LUT12 import TrackSegmentFinderLUT12
from b2tsf.track_segment_finder_hourglas import TrackSegmentFinderHourglas

class TrackSegmentFinder:
    """
    Class TrackSegmentFinder instantiates any necessary track segment
    finders and returns them to the end user. Multiple functions are hidden
    behind generic caller functions and type of function call or track
    segment finder is defined by mode parameter.
    ---------------------------------------------------------------------------
    Possible modes for constructor:
    ---------------------------------------------------------------------------

    """
    def __init__(self,mode,**kwargs):
        """
        mode='hourglas':   Baseline algorithm implemented now in Belle II.

        mode='LUT':     Baseline algorithm implementing 5 bit bin look up tables.

        mode='LUT9':    Algorithm implementing 9 bit bin look up tables.

        mode='LUT12':   Algorithm implementing 12 bit bin look up tables.

        mode='Neuro':   Not yet implemented
        """
        if(mode == 'hourglas'):
            self._tsf = TrackSegmentFinderHourglas()
        elif(mode == 'LUT5'):
            self._tsf = TrackSegmentFinderLUT5()
        elif(mode == 'LUT9'):
            self._tsf = TrackSegmentFinderLUT9()
        elif(mode == 'LUT12'):
            self._tsf = TrackSegmentFinderLUT12()
        else:
            raise ValueError(mode)
        self.mode = mode


    def set_patterns(self,binmode,patterns_lower,patterns_upper):
        """
        Callable from track segment finder in mode 'LUT9'. Throws warning
        otherwise. Gives access to dictionary setting of LUT9 track segement
        finder.
        """
        if(self.mode == 'LUT9'):
            self._tsf = self._tsf.set_patterns(binmode,
                                               patterns_lower,
                                               patterns_upper)
        elif(self.mode == 'LUT12'):
            self._tsf = self._tsf.set_patterns(binmode,
                                               patterns_lower,
                                               patterns_upper)
        else:
            warnings.warn("This TSF mode can not be configured.")
        return self


    def get_hit_list(self,format='native'):
        """
        Returns list of hits, e.g. datapoints of internal cdc frame. Data format
        can be choosen between 'native' and 'r_theta'.
        """
        return self._tsf.get_frame().get_hit_list(format)


    def get_number_of_hits(self):
        """
        Returns number of hits. Throws warning if mode is not set to 'hits' for
        track segment finder. In this case number of possible hit map hits would be
        reported.
        """
        n_hits = self._tsf.get_frame().get_number_of_hits()

        if(self._tsf.mode == 'hits'):
            pass
        else:
            warnings.warn("TSF mode is not set to 'hits'. This can lead to unwanted behavior.")

        return n_hits


    def detect(self,frame_in,mode='hits',edge_hit=True):
        """
        Detect hits and returns internal frame as out frame.
        """
        return self._tsf.detect(frame_in,mode,edge_hit)


    def get_segment_dword(self,segment_id):
        """
        Returns the output of one processed segment by natively given out a 32
        list with '0' or '1' whether the respective hit bin has been hit. For
        data format description see thesis.
        """
        return self._tsf.get_segment_dword(segment_id)


    def hit_compaction(self):
        """
        Performs stream compaction algorithm and returns resulting list of tsf
        hit ids. List for each layer returns up to n_tsout elements.
        """
        return self._tsf.hit_compaction()


    def get_n_tsout(self):
        """
        Returns hardware design parameter n_tsout. This resembles the maximum
        number of tsf hits that can be sent per layer per hw clock cycle.
        """
        if(self.mode == 'LUT5'):
            return self._tsf.get_n_tsout()
        elif(self.mode == 'LUT9'):
            return self._tsf.get_n_tsout()
        elif(self.mode == 'LUT12'):
            return self._tsf.get_n_tsout()
        else:
            ValueError(f"Function get_n_tsout not avaible for mode {self.mode}")


    def clear(self):
        """
        Clears data without recreating the TrackSegmentFinder object.
        """
        return self._tsf.clear()



from datetime import date


class TSFProcessorRefTestbenchWriter:
    """
    TSFProcessorTestbenchWriter reads in a sequence of tsf hit lists and writes
    them to a specific path.
    ----------------------------------------------------------------------------
    Methods:
    ----------------------------------------------------------------------------
    set_sequence(data)
        Accepts 3rd order list of order hit_list = data[sequence_step,layer_id]
        as parameter <data>.

    write_layer(path,filename,layer_id,n_hits)
        Writes data defined with set_sequence before to a file named filename
        under path, writing only layer with layer_id and filling up hit_lists
        with 0s up to n_hits.
    """

    def __init__(self):
        """
        Basic setup of textual parameters.
        """
        self.mode = 'tsf_processor_ref'
        self.offset = 0
        #TODO: Retrieve from Package
        self.version = 0.1
        self.date = '{:%Y-%m-%d}'.format(date.today())
        self.author = 'Marc Neu'
        self._intro_text = "#FileWriter V. {version} generating a {mode} file\n#\n#Author: {author}\n#Date: {date}\n#\n".format(version=self.version,date=self.date,author=self.author,mode=self.mode)
        self._outro_text = "#EOF."


    def set_sequence(self, data):
        """
        Accepts 3rd order list of order hit_list = data[sequence_step,layer_id]
        as parameter <data>.
        """
        self._data = data
        return self


    def set_offset(self, offset):
        """
        Sets delay of output lines in the model checker
        compared to the actual stimuli values.
        Can be used to model latency.
        Empty lines are filled with 0's.
        """
        self._offset = offset


    def _gen_line(self,hit_ids,n_hits):
        """
        Reads tsf output data and creates a string in a predefined format
        (see thesis)
        """
        hit_list = hit_ids.copy()
        hit_list.reverse()
        hit_list = [0] * (n_hits - len(hit_ids)) + hit_list
        reference_string = ''.join(f'{item:0>10b}' for item in hit_list)

        return reference_string


    def _write_layer(self,path,filename,layer_id,n_hits):
        """
        Writes a all segments of the specified layer to a file under <path>
        """
        file_path = "{path}{filename}_{id:d}.dat".format(path=path,filename=filename,id=layer_id)

        with open(file_path,"w") as file_ptr:
            file_ptr.write(self._intro_text)

            for i in range(self._offset):
                file_ptr.write(f":empty\n")

            for sim_frame in self._data:
                hit_ids = sim_frame[layer_id]
                data_line = self._gen_line(hit_ids,n_hits)
                file_ptr.write(data_line + "\n")

            file_ptr.write(self._outro_text)

    def write_layer(self,path,filename,layer_id,n_hits):
        """
        Writes data defined with set_sequence before to a file named filename
        under path, writing only layer with layer_id and filling up hit_lists
        with 0s up to n_hits.
        """
        self._write_layer(path,filename,layer_id,n_hits)



import numpy
from bitarray import bitarray
from bitarray import frozenbitarray
import warnings
from b2tsf.belle_CDC_frame import CDCFrame

class TrackSegmentFinderLUT12:
    """
    Track Segement Finder implementing a 12 bit look up table solution.

    ---------------------------------------------------------------------------
    Methods:
    ---------------------------------------------------------------------------
    detect:     Detects hits in incoming frame <frame_in> and saves result in
                internal data strucutre. Two possible detetction modes are
                available. 'hits' returns real hardware hits whereas 'candidates'
                returns all hit candidates of a hit bin.

    get_frame:  Retruns internal data frame as reference.

    clear:      Clears internal data structure.

    """
    def __init__(self):
        self._frame_out = CDCFrame(0)
        self._binsize = 12
        self._binnumber = 16
        self._binlayer = 2
        self.mode = 'hits'

        self._init = False
        self._frame_in = None
        self._frame_in_delayed = None
        self._cooldown_array = None
        self._cooldown = 64
        self._n_tsout = 24


    def _init_patterns(self,patterns):
        """
        Configure detection patterns during track segment finder construction. True
        candidate patterns are given as string list in the format defined in thesis
        documenation. As keys must be constant to be hashable this function must be
        be called seperately.
        Returns dict of bit patterns.
        """
        keys = []
        values = []
        for i in range(2**self._binsize):
            key_pattern = f'{i:012b}'
            keys.append(frozenbitarray(key_pattern))
            values.append(key_pattern in patterns)
        return dict(zip(keys,values))


    def _detect_hit(self,hitmap_zip,seg_hits):
        """
        Implementation of per bin detetion algorithm, e.g. checking against
        look up table in form of a hash map. hitmap_zip contains both bins
        for upper and lower hitmap.
        This function requires a predefined array for performance reasons. This
        way the creation of many small arrays can be avoided by slicing.
        """
        hitmap_lower, hitmap_upper = hitmap_zip
        if(self._binmode == 'lower' or self._binmode == 'both'):
            for i in range(self._binnumber):
                seg_hits[0,i] = self._patterns_lower[hitmap_lower[i]]
        if(self._binmode == 'upper' or self._binmode == 'both'):
            for i in range(self._binnumber):
                seg_hits[1,i] = self._patterns_upper[hitmap_upper[i]]
        return seg_hits


    def _populate_bin(self,seg_pre, seg_act, seg_suc, edge_hit):
        """
        Populate bins for detection algorithm. Each segment has got 16 bins in 2 layers
        as specified in the alogrithm description in thesis. Furthermore edge_hits of
        sectors can be toggeled on/off.
        This function returns bins of a layer segment.
        """
        hit_bin_lower = [self._binsize * bitarray() for i in range(self._binnumber)]
        hit_bin_upper = [self._binsize * bitarray() for i in range(self._binnumber)]
        #Fill lower bins
        if(self._binmode == 'lower' or self._binmode == 'both'):
            if(edge_hit):
                hit_bin_lower[0] = frozenbitarray([
                    seg_act[2,1],seg_act[2,0],seg_pre[2,15],
                    seg_act[1,1],seg_act[1,0],seg_pre[1,15],seg_pre[1,14],
                    seg_act[0,2],seg_act[0,1],seg_act[0,0],seg_pre[0,15],seg_pre[0,14]])
                hit_bin_lower[1] = frozenbitarray([
                    seg_act[2,2],seg_act[2,1],seg_act[2,0],
                    seg_act[1,2],seg_act[1,1],seg_act[1,0],seg_pre[1,15],
                    seg_act[0,3],seg_act[0,2],seg_act[0,1],seg_act[0,0],seg_pre[0,15]])
                hit_bin_lower[14] = frozenbitarray([
                    seg_act[2,15],seg_act[2,14],seg_act[2,13],
                    seg_act[1,15],seg_act[1,14],seg_act[1,13],seg_act[1,12],
                    seg_suc[0,0],seg_act[0,15],seg_act[0,14],seg_act[0,13],seg_act[0,12]])
                hit_bin_lower[15] = frozenbitarray([
                    seg_suc[2,0],seg_act[2,15],seg_act[2,14],
                    seg_suc[1,0],seg_act[1,15],seg_act[1,14],seg_act[1,13],
                    seg_suc[0,1],seg_suc[0,0],seg_act[0,15],seg_act[0,14],seg_act[0,13]])
            else:
                hit_bin_lower[0] = frozenbitarray([
                    seg_act[2,1],seg_act[2,0],0,
                    seg_act[1,1],seg_act[1,0],0,0,
                    seg_act[0,2],seg_act[0,1],seg_act[0,0],0,0])
                hit_bin_lower[1] = frozenbitarray([
                    seg_act[2,2],seg_act[2,1],seg_act[2,0],
                    seg_act[1,2],seg_act[1,1],seg_act[1,0],0,
                    seg_act[0,3],seg_act[0,2],seg_act[0,1],seg_act[0,0],0])
                hit_bin_lower[14] = frozenbitarray([
                    seg_act[2,15],seg_act[2,14],seg_act[2,13],
                    seg_act[1,15],seg_act[1,14],seg_act[1,13],seg_act[1,12],
                    0,seg_act[0,15],seg_act[0,14],seg_act[0,13],seg_act[0,12]])
                hit_bin_lower[15] = frozenbitarray([
                    0,seg_act[2,15],seg_act[2,14],
                    0,seg_act[1,15],seg_act[1,14],seg_act[1,13],
                    0,0,seg_act[0,15],seg_act[0,14],seg_act[0,13]])

            for i in range(0+2,self._binnumber-2):
                hit_bin_lower[i] = frozenbitarray([
                    seg_act[2,i+1],seg_act[2,i],seg_act[2,i-1],
                    seg_act[1,i+1],seg_act[1,i],seg_act[1,i-1],seg_act[1,i-2],
                    seg_act[0,i+2],seg_act[0,i+1],seg_act[0,i],seg_act[0,i-1],seg_act[0,i-2]])
        #Fill upper bins
        if(self._binmode == 'upper' or self._binmode == 'both'):
            if(edge_hit):
                hit_bin_upper[0] = frozenbitarray([
                    seg_act[4,2],seg_act[4,1],seg_act[4,0],seg_pre[4,15],seg_pre[4,14],
                    seg_act[3,1],seg_act[3,0],seg_pre[3,15],seg_pre[3,14],
                    seg_act[2,1],seg_act[2,0],seg_pre[2,15]])
                hit_bin_upper[1] = frozenbitarray([
                    seg_act[4,3],seg_act[4,2],seg_act[4,1],seg_act[4,0],seg_pre[4,15],
                    seg_act[3,2],seg_act[3,1],seg_act[3,0],seg_pre[3,15],
                    seg_act[2,2],seg_act[2,1],seg_act[2,0]])
                hit_bin_upper[14] = frozenbitarray([
                    seg_suc[4,0],seg_act[4,15],seg_act[4,14],seg_act[4,13],seg_act[4,12],
                    seg_act[3,15],seg_act[3,14],seg_act[3,13],seg_act[3,12],
                    seg_act[2,15],seg_act[2,14],seg_act[2,13]])
                hit_bin_upper[15] = frozenbitarray([
                    seg_suc[4,1],seg_suc[4,0],seg_act[4,15],seg_act[4,14],seg_act[4,13],
                    seg_suc[3,0],seg_act[3,15],seg_act[3,14],seg_act[4,13],
                    seg_suc[2,0],seg_act[2,15],seg_act[2,14]])
            else:
                hit_bin_upper[0] = frozenbitarray([
                    seg_act[4,2],seg_act[4,1],seg_act[4,0],0,0,
                    seg_act[3,1],seg_act[3,0],0,0,
                    seg_act[2,1],seg_act[2,0],0])
                hit_bin_upper[1] = frozenbitarray([
                    seg_act[4,3],seg_act[4,2],seg_act[4,1],seg_act[4,0],0,
                    seg_act[3,2],seg_act[3,1],seg_act[3,0],0,
                    seg_act[2,2],seg_act[2,1],seg_act[2,0]])
                hit_bin_upper[14] = frozenbitarray([
                    0,seg_act[4,15],seg_act[4,14],seg_act[4,13],seg_act[4,12],
                    seg_act[3,15],seg_act[3,14],seg_act[3,13],seg_act[3,12],
                    seg_act[2,15],seg_act[2,14],seg_act[2,13]])
                hit_bin_upper[15] = frozenbitarray([
                    0,0,seg_act[4,15],seg_act[4,14],seg_act[4,13],
                    0,seg_act[3,15],seg_act[3,14],seg_act[4,13],
                    0,seg_act[2,15],seg_act[2,14]])

            for i in range(0+2,self._binnumber-2):
                hit_bin_upper[i] = frozenbitarray([
                    seg_act[4,i+2],seg_act[4,i+1],seg_act[4,i],seg_act[4,i-1],seg_act[4,i-2],
                    seg_act[3,i+1],seg_act[3,i],seg_act[3,i-1],seg_act[3,i-2],
                    seg_act[2,i+1],seg_act[2,i],seg_act[2,i-1]])

        return (hit_bin_lower,hit_bin_upper)


    def _check_layer_hit(self,layer_in, edge_hit):
        """
        Checks frame layer <layer_in> for hits. Using edge_hit True/False segment edge mode can be
        choosen. For production use edge_hit=True as non edge_hit is only used for hardware module
        testing.
        """
        n_segs = layer_in.n_segs
        hits = numpy.zeros(shape=(n_segs,self._binlayer,self._binnumber),dtype=int)
        #Segment 0
        seg_pre = layer_in.get_segment(n_segs-1)
        seg_act = layer_in.get_segment(0)
        seg_suc = layer_in.get_segment(1)
        hit_bin = self._populate_bin(seg_pre, seg_act, seg_suc,edge_hit)
        hits[0,:,:] = self._detect_hit(hit_bin, hits[0,:,:])
        #Inner segments
        for i in range (1,n_segs-1):
            seg_pre = layer_in.get_segment(i-1)
            seg_act = layer_in.get_segment(i)
            seg_suc = layer_in.get_segment(i+1)
            hit_bin = self._populate_bin(seg_pre, seg_act, seg_suc,edge_hit)
            hits[i,:,:] = self._detect_hit(hit_bin, hits[i,:,:])
        #Segment n_seg
        seg_pre = layer_in.get_segment(n_segs-2)
        seg_act = layer_in.get_segment(n_segs-1)
        seg_suc = layer_in.get_segment(0)
        hit_bin = self._populate_bin(seg_pre, seg_act, seg_suc,edge_hit)
        hits[-1,:,:] = self._detect_hit(hit_bin, hits[-1,:,:])

        return hits

    def _detect_hits(self,frame_in,edge_hit=True):
        """
        Detects hits of all layers in <frame_in> and saves output in tsf frame.
        Hardcoded to layer 1...8, because hardware implementation for layer 0 is
        non existent. This feature could be implemented later.

        Step 1: Calculate hits for each layer. Returns 2D Array of hits
                with indices <segment_id> whose range depends on the layer
                and 0...31 for each checked hit bin inside the dection algorithm.
                E.g. Bin 5 in Segment 3 was hit => hits[2,4] == True

        Step 2: Find all hits and fill output frame accordingly.
        """
        self._frame_in = frame_in
        for layer_id in range(1,9):
            layer_in = frame_in.get_layer(layer_id)
            layer_segments = self._frame_out.get_layer(layer_id).get_tsf_segments()
            #Calculate hits of layer
            hits = self._check_layer_hit(layer_in,edge_hit)
            (seg_id_list,row_id_list,col_id_list) = numpy.nonzero(hits)
            for seg_id,up_lo_id,col_id in zip(seg_id_list,row_id_list,col_id_list):
                #Remove if row_id
                layer_segments[seg_id,0+4*up_lo_id,col_id] = 1
        return self._frame_out

    def _detect_candidates(self,frame_in,edge_hit=True):
        """
        Detects hits of all layers in <frame_in> and saves output in tsf frame.
        Hardcoded to layer 1...8, because hardware implementation for layer 0 is
        non existent. This feature could be implemented later.

        Step 1: Calculate hits for each layer. Returns 2D Array of hits
                with indices <segment_id> whose range depends on the layer
                and 0...31 for each checked hit bin inside the dection algorithm.
                E.g. Bin 5 in Segment 3 was hit => hits[2,4] == True

        Step 2: Find all hits and fill output frame accordingly.

        Identical to _detect_hits() except that it populates all hit bin
        candidates to the internal output data frame.
        """
        self._frame_in = frame_in
        for layer_id in range(1,9):
            layer_in = frame_in.get_layer(layer_id)
            layer_segments = self._frame_out.get_layer(layer_id).get_tsf_segments()
            #Calculate hits of layer
            hits = self._check_layer_hit(layer_in,edge_hit)
            (seg_id_list,up_lo_id,col_id_list) = numpy.nonzero(hits)
            for seg_id,up_lo_id,col_id in zip(seg_id_list,up_lo_id,col_id_list):
                if(up_lo_id == 0):
                    if(col_id == 0):
                        layer_segments[seg_id-1,0,14:] = 1
                        layer_segments[seg_id,0,:3] = 1
                        layer_segments[seg_id-1,1,14:] = 1
                        layer_segments[seg_id,1,:2] = 1
                        layer_segments[seg_id-1,2,15] = 1
                        layer_segments[seg_id,2,:2] = 1
                    elif(col_id == 1):
                        layer_segments[seg_id-1,0,15] = 1
                        layer_segments[seg_id,0,:4] = 1
                        layer_segments[seg_id-1,1,15] = 1
                        layer_segments[seg_id,1,:3] = 1
                        layer_segments[seg_id,2,:3] = 1
                    elif(col_id == 14):
                        layer_segments[seg_id,0,12:] = 1
                        layer_segments[(seg_id+1)%layer_in.n_segs,0,0] = 1
                        layer_segments[seg_id,1,12:] = 1
                        layer_segments[seg_id,2,13:] = 1
                    elif(col_id == 15):
                        layer_segments[seg_id,0,13:] = 1
                        layer_segments[(seg_id+1)%layer_in.n_segs,0,:2] = 1
                        layer_segments[seg_id,1,13:] = 1
                        layer_segments[(seg_id+1)%layer_in.n_segs,1,0] = 1
                        layer_segments[seg_id,2,14:] = 1
                        layer_segments[(seg_id+1)%layer_in.n_segs,2,0] = 1
                    else:
                        layer_segments[seg_id,0,col_id-2:col_id+3] = 1
                        layer_segments[seg_id,1,col_id-2:col_id+2] = 1
                        layer_segments[seg_id,2,col_id-1:col_id+2] = 1
                if(up_lo_id == 1):
                    if(col_id == 0):
                        layer_segments[seg_id-1,4,14:] = 1
                        layer_segments[seg_id,4,0:3] = 1
                        layer_segments[seg_id-1,3,14:] = 1
                        layer_segments[seg_id,3,0:2] = 1
                        layer_segments[seg_id-1,2,15] = 1
                        layer_segments[seg_id,2,0:2] = 1
                    elif(col_id == 1):
                        layer_segments[seg_id-1,4,15] = 1
                        layer_segments[seg_id,4,0:4] = 1
                        layer_segments[seg_id-1,3,15] = 1
                        layer_segments[seg_id,3,0:3] = 1
                        layer_segments[seg_id,2,0:3] = 1
                    elif(col_id == 14):
                        layer_segments[seg_id,4,12:] = 1
                        layer_segments[(seg_id+1)%layer_in.n_segs,4,0] = 1
                        layer_segments[seg_id,3,12:] = 1
                        layer_segments[seg_id,2,13:] = 1
                    elif(col_id == 15):
                        layer_segments[seg_id,4,13:] = 1
                        layer_segments[(seg_id+1)%layer_in.n_segs,4,0:2] = 1
                        layer_segments[seg_id,3,13:] = 1
                        layer_segments[(seg_id+1)%layer_in.n_segs,3,0] = 1
                        layer_segments[seg_id,2,14:] = 1
                        layer_segments[(seg_id+1)%layer_in.n_segs,2,0] = 1
                    else:
                        layer_segments[seg_id,4,col_id-2:col_id+3] = 1
                        layer_segments[seg_id,3,col_id-2:col_id+2] = 1
                        layer_segments[seg_id,2,col_id-1:col_id+2] = 1
        return self._frame_out


    def _detect_hw_model(self,frame_in,edge_hit=True):
        """
        Detects hit in a sequence, if Track Segment Finder is called multiple
        times. Returns frame with hit data.
        """
        if(not self._init):
            self._frame_in_delayed = CDCFrame(0)
            #24 is the maximum number of segments, e.g. n_segs in layer 9.
            self._cooldown_array = numpy.zeros(shape=(9,
                                                      24,
                                                      self._binlayer,
                                                      self._binnumber),
                                               dtype=int)
            self._init = True
        else:
            self._frame_in_delayed = self._frame_in_delayed.copy(self._frame_in)
            self._cooldown_array = numpy.where(self._cooldown_array == 0,
                                                self._cooldown_array,
                                                self._cooldown_array-1)
        self._frame_in = frame_in
        frame_work = self._frame_in_delayed.logical_or(frame_in)

        self._frame_out.clear()

        for layer_id in range(1,9):
            #frame_work
            layer_in = frame_work.get_layer(layer_id)
            layer_segments = self._frame_out.get_layer(layer_id).get_tsf_segments()
            #Calculate hits of layer
            hits = self._check_layer_hit(layer_in,edge_hit)
            (seg_id_list,row_id_list,col_id_list) = numpy.nonzero(hits)
            for seg_id,up_lo_id,col_id in zip(seg_id_list,row_id_list,col_id_list):
                if(self._cooldown_array[layer_id,seg_id,up_lo_id,col_id] == 0):
                    self._cooldown_array[layer_id,seg_id,up_lo_id,col_id] = self._cooldown
                    layer_segments[seg_id,0+up_lo_id*4,col_id] = 1
        return self._frame_out


    def get_frame(self):
        return self._frame_out


    def get_n_tsout(self):
        return self._n_tsout


    def set_patterns(self,binmode,patterns_lower,patterns_upper):
        """
        Creates patterns for lower and upper hit bins. Caution: This function is
        computational expensive!
        """
        self._binmode = binmode
        if(binmode != 'lower' and binmode != 'upper' and binmode != 'both'):
            raise ValueError(binmode)
        self._patterns_lower = self._init_patterns(patterns_lower)
        self._patterns_upper = self._init_patterns(patterns_upper)
        return self


    def detect(self,frame_in,mode='hits',edge_hit=True):
        """
        Interface function for detection algorithm.
        """
        self.mode = mode
        if(mode == 'hits'):
            return self._detect_hits(frame_in,edge_hit)
        elif(mode == 'candidates'):
            return self._detect_candidates(frame_in,edge_hit)
        elif(mode == 'hw_model'):
            return self._detect_hw_model(frame_in,edge_hit)
        else:
            raise ValueError('Mode: ' + mode + 'does not exist. See documentation for valid arguments.')

    def get_segment_dword(self,segment_id):
        """
        Returns list of segment hits as described in thesis.
        """
        dword_layer_list = []
        for work_layer in self._frame_out.get_layers():
            work_seg = work_layer.get_tsf_segment(segment_id)

            dword = numpy.zeros(self._binlayer*self._binnumber,dtype=int)
            dword[:self._binnumber] = work_seg[0,:]
            dword[self._binnumber:self._binnumber*self._binlayer] = work_seg[4,:]

            dword_layer_list.append(dword.tolist())

        return dword_layer_list


    def hit_compaction(self):
        """
        Returns compact list of hits, models stream compaction implementation.
        See master thesis for description of id sorting.
        Works on internal frame which is poulated by callind detect()
        beforehand.
        Returns:
            List of hit list of layers 0 to 9.
        """
        (layer_id_list,
         seg_id_list,
         row_id_list,
         col_id_list) = self._frame_out.get_hit_list(format='native',
                                                      detect_hit=True,
                                                      skip_inner_layer=True)
        hit_list = [[] for i in range(0,9)]

        for (layer_id,seg_id,row_id,col_id) in zip(layer_id_list,seg_id_list,row_id_list,col_id_list):
            if(row_id < 2):
                hardware_id = seg_id*self._binnumber*self._binlayer + col_id + 1
            else:
                hardware_id = seg_id*self._binnumber*self._binlayer + col_id + self._binnumber + 1
            hit_list[layer_id].append(hardware_id)


        #Hardware is only able to process n_tsout number of track segemnts. All other segments
        #are discarded

        hit_list = [hit_list[i][:self._n_tsout] for i in range(0,9)]

        return hit_list


    def clear(self):
        """
        Clears track segment finder internal files. Resets hardware sequential
        model. Returns empty output frame.
        """
        self._frame_out = self._frame_out.clear()
        self._init = False
        return self._frame_out



from b2tsf.belle_CDC_frame import CDCFrame

class CDCSequence:
    """
    Sequence of CDC data frames in time domain.

    ---------------------------------------------------------------------------
    Methods:
    ---------------------------------------------------------------------------

    """

    def __init__(self,frame_list=None,t_step=1):
        if(frame_list is None):
            self._frame_list = []
            self._n_steps = 0
        else:
            self._frame_list = frame_list
            self._n_steps = len(frame_list)
        self._t_step = t_step
        self._t_min = 0
        self._t_max = self._t_step*self._n_steps


    def get_sequence_time_range(self):
        return (self._t_min,self._t_max)


    def get_sequence_step_range(self):
        return (0, self._n_steps)


    def get_frame_by_id(self,id):
        return self._frame_list[id]


    def get_frames(self):
        return self._frame_list


    def append_frames(self,frames):
        self._frame_list.extend(frames)
        self._n_steps = self._n_steps + len(frames)
        self._t_max = self._t_step*self._n_steps

    def append_frame(self,frame):
        self._frame_list.append(frame)
        self._n_steps = self._n_steps + 1
        self._t_max = self._t_step*self._n_steps

    def append_empty(self,n=1):
        for i in range(n):
            self._frame_list.append(CDCFrame())
            self._n_steps = self._n_steps + 1
            self._t_max = self._t_step*self._n_steps

    def append_full(self,n=1):
        for i in range(n):
            self._frame_list.append(CDCFrame().fill())
            self._n_steps = self._n_steps + 1
            self._t_max = self._t_step*self._n_steps



from b2tsf.tsf_processor_ref_testbench_writer import TSFProcessorRefTestbenchWriter
from b2tsf.tsf_processor_stimuli_testbench_writer import TSFProcessorStimuliTestbenchWriter
from b2tsf.merger_processor_ref_testbench_writer import MergerProcessorRefTestbenchWriter
from b2tsf.merger_processor_stimuli_testbench_writer import MergerProcessorStimuliTestbenchWriter

class TestbenchWriterFactory:
    """
    Factory class for independend use of TestbenchWriters. Mode defines the type
    of the writer as in
    mode == 'tsf_processor_ref'

    mode == 'merger_processor_ref'

    mode == 'tsf_processor_stimuli'

    mode == 'merger_processor_stimuli'
    """

    def __init__(self):
       pass


    def get_writer(self,mode='tsf_processor_stimuli'):
        if(mode == 'tsf_processor_ref'):
            return TSFProcessorRefTestbenchWriter()
        elif(mode == 'merger_processor_ref'):
            return MergerProcessorRefTestbenchWriter()
        elif(mode == 'tsf_processor_stimuli'):
            return TSFProcessorStimuliTestbenchWriter()
        elif(mode == 'merger_processor_stimuli'):
            return MergerProcessorStimuliTestbenchWriter()
        else:
            ValueError("TestbenchWriter of mode" + mode + "is unknown.")



import pkg_resources
import numpy
import warnings
import array
import plotly.graph_objects as go
from b2tsf.belle_CDC_layer import CDCLayer
from b2tsf.belle_CDC_frame import CDCFrame

class TrackSegmentFinderHourglas:
    """
    Track Segement Finder implenmenting the conventional hourglas solution.

    ---------------------------------------------------------------------------
    Methods:
    ---------------------------------------------------------------------------
    detect:     Detects hits in incoming frame <frame_in> and saves result in
                internal data strucutre. Two possible detetction modes are
                available. 'hits' returns real hardware hits whereas 'candidates'
                returns all hit candidates of a hit bin.

    get_frame:  Retruns internal data frame as reference.

    clear:      Clears internal data structure.

    """

    def __init__(self):
        """
        A priori probabilities are saved in a txt look up table and loaded
        during object creation. Make sure the path is provided accordingly.
        """

        self._frame = CDCFrame(0)
        self.mode = 'hits'
        self._outer_lr_lut = numpy.loadtxt(pkg_resources.resource_filename(__name__, 'data/outerLRLUT.mif'),
                                           comments="#",
                                           delimiter=",",
                                           converters={0: lambda x: int(x,2)},
                                           unpack=False)
        self._inner_lr_lut = numpy.loadtxt(pkg_resources.resource_filename(__name__, 'data/innerLRLUT.mif'),
                                           comments="#",
                                           delimiter=",",
                                           converters={0: lambda x: int(x,2)},
                                           unpack=False)

    def _detect_hit_inner(self,hitmap):
        """
        Hitmap is an encoded 1D array with 16 elements.
        Elements 0-A are hits at respective addresses.
        Element B is buffer element an therefore 0.
        Line Number |     Hits     |   Address
        ------------|--------------|-------------
            5       | -X-X-X-X-X-- | -0-1-2-3-4--
            4       | --X-X-X-X-O- | --5-6-7-8---
            3       | -O-X-X-X-O-- | ---9-A-B----
            2       | --O-X-X-O-O- | ----C-D-----
            1       | -O-O-X-O-O-- | -----E------
        ------------|--------------|-------------
        As defined in CDC Merger Board Specification v1.
        """
        #TODO: Improve this code for efficiency and readability.
        #It is True only when first priority
        #cell is not hit and the left secondary one is hit.
        hitmap[15] = int((not hitmap[14]) and hitmap[12])
        #Calculate Look-Up-Table address
        #Use hitmap array, convert to string and convert string to integer
        lr_addr = int(''.join(map(str,hitmap)),2)
        lr = self._inner_lr_lut[lr_addr]

        hitline5 = hitmap[0] or hitmap[1] or hitmap[2] or hitmap[3] or hitmap[4]
        hitline4 = hitmap[5] or hitmap[6] or hitmap[7] or hitmap[8]
        hitline3 = hitmap[9] or hitmap[10] or hitmap[11]
        hitline2 = hitmap[12] or hitmap[13]
        hitline1 = hitmap[14]

        result = (hitline4 and hitline3 and hitline2 and hitline1 and  lr) or \
                 (hitline5 and hitline3 and hitline2 and hitline1 and  lr) or \
                 (hitline5 and hitline4 and hitline3 and hitline1 and  lr) or \
                 (hitline5 and hitline4 and hitline3 and hitline2 and  lr) or \
                 (hitline5 and hitline4 and hitline3 and hitline2 and  lr) or \
                 (hitline5 and hitline4 and hitline3 and hitline2 and hitline1 and lr)

        return result


    def _detect_hit_outer(self,hitmap):
        """
        Hitmap is an encoded 1D array with 12 elements.
        Elements 0-A are hits at respective addresses.
        Element B is buffer element an therefore 0.
        Line Number |     Hits     |   Address
        ------------|--------------|-------------
            5       | -X-O-O-O-X-- | ---0-1-2----
            4       | --X-O-O-X-X- | ----3-4-----
            3       | -X-X-O-X-X-- | -----5------
            2       | --X-O-O-X-X- | ----6-7-----
            1       | -X-O-O-O-X-- | ---8-9-A----
        ------------|--------------|-------------
        As defined in CDC Merger Board Specification v1.
        """
        #TODO: Improve this code for efficiency and readability.
        #It is True only when first priority
        #cell is not hit and the left secondary one is hit.
        hitmap[11] = int((not hitmap[5]) and hitmap[3])
        #Calculate Look-Up-Table address
        #Use hitmap array, convert to string and convert string to integer
        lr_addr = int(''.join(map(str,hitmap)),2)
        lr = self._outer_lr_lut[lr_addr]

        hitline5 = hitmap[0] or hitmap[1] or hitmap[2]
        hitline4 = hitmap[3] or hitmap[4]
        hitline2 = hitmap[6] or hitmap[7]
        hitline1 = hitmap[8] or hitmap[9] or hitmap[10]

        result = (hitmap[5] and lr) or \
                 (hitline5 and hitline4 and hitline2 and hitline1 and lr)

        return result

    def _populate_bin_inner(self,seg_pre, seg_act, seg_suc, edge_hit):
        """
        Populate bins for detection algorithm. Each segment has got 16 bins as specified.
        Furthermore edge_hits of sectors can be toggeled on/off.
        This function returns bins of a layer segment.
        """
        hit_bin = 16 * [None]

        if(edge_hit):
            # Circulation counterclockwise - successor left, predecessor right
            # Add 0 element as placeholder for lr value
            hit_bin[0] = list(map(int,[seg_act[4,2],seg_act[4,1],seg_act[4,0],seg_pre[4,15],seg_pre[4,14],
                                       seg_act[3,2],seg_act[3,1],seg_act[3,0],seg_pre[3,15],
                                       seg_act[2,1],seg_act[2,0],seg_pre[2,15],
                                       seg_act[1,1],seg_act[1,0],
                                       seg_act[0,0],
                                       0]))
            hit_bin[1] = list(map(int,[seg_act[4,3],seg_act[4,2],seg_act[4,1],seg_act[4,0],seg_pre[4,15],
                                       seg_act[3,3],seg_act[3,2],seg_act[3,1],seg_act[3,0],
                                       seg_act[2,2],seg_act[2,1],seg_act[2,0],
                                       seg_act[1,2],seg_act[1,1],
                                       seg_act[0,1],
                                       0]))
            hit_bin[14] =  list(map(int,[seg_suc[4,0],seg_act[4,15],seg_act[4,14],seg_act[4,13],seg_act[4,12],
                                         seg_suc[3,0],seg_act[3,15],seg_act[3,14],seg_act[3,13],
                                         seg_act[2,15],seg_act[2,14],seg_act[2,13],
                                         seg_act[1,15],seg_act[1,14],
                                         seg_act[0,14],
                                         0]))
            hit_bin[15] =  list(map(int,[seg_suc[4,1],seg_suc[4,0],seg_act[4,15],seg_act[4,14],seg_act[4,13],
                                         seg_suc[3,1],seg_suc[3,0],seg_act[3,15],seg_act[3,14],
                                         seg_suc[2,0],seg_act[2,15],seg_act[2,14],
                                         seg_suc[1,0],seg_act[1,15],
                                         seg_act[0,15],
                                         0]))
        else:
            hit_bin[0] = list(map(int,[seg_act[4,2],seg_act[4,1],seg_act[4,0],0,0,
                                       seg_act[3,2],seg_act[3,1],seg_act[3,0],0,
                                       seg_act[2,1],seg_act[2,0],0,
                                       seg_act[1,1],seg_act[1,0],
                                       seg_act[0,0],
                                       0]))
            hit_bin[1] = list(map(int,[seg_act[4,3],seg_act[4,2],seg_act[4,1],seg_act[4,0],0,
                                       seg_act[3,3],seg_act[3,2],seg_act[3,1],seg_act[3,0],
                                       seg_act[2,2],seg_act[2,1],seg_act[2,0],
                                       seg_act[1,2],seg_act[1,1],
                                       seg_act[0,1],
                                       0]))
            hit_bin[14] =  list(map(int,[0,seg_act[4,15],seg_act[4,14],seg_act[4,13],seg_act[4,12],
                                         0,seg_act[3,15],seg_act[3,14],seg_act[3,13],
                                         seg_act[2,15],seg_act[2,14],seg_act[2,13],
                                         seg_act[1,15],seg_act[1,14],
                                         seg_act[0,14],
                                         0]))
            hit_bin[15] =  list(map(int,[0,0,seg_act[4,15],seg_act[4,14],seg_act[4,13],
                                         0,0,seg_act[3,15],seg_act[3,14],
                                         0,seg_act[2,15],seg_act[2,14],
                                         0,seg_act[1,15],
                                         seg_act[0,15],
                                         0]))

        for i in range(2,16-2):
            hit_bin[i] =  list(map(int,[seg_act[4,i+2],seg_act[4,i+1],seg_act[4,i],seg_act[4,i-1],seg_act[4,i-2],
                                        seg_act[3,i+2],seg_act[3,i+1],seg_act[3,i],seg_act[3,i-1],
                                        seg_act[2,i+1],seg_act[2,i],seg_act[2,i-1],
                                        seg_act[1,i+1],seg_act[1,i],
                                        seg_act[0,i],
                                        0]))
        return hit_bin


    def _populate_bin_outer(self,seg_pre, seg_act, seg_suc, edge_hit):
        """
        Populate bins for detection algorithm. Each segment has got 16 bins as specified.
        Furthermore edge_hits of sectors can be toggeled on/off.
        This function returns bins of a layer segment.
        """
        hit_bin = 16 * [None]

        if(edge_hit):
            # Circulation counterclockwise - successor left, predecessor right
            # Add 0 element as placeholder for lr value
            hit_bin[0] = list(map(int,[seg_act[4,1],seg_act[4,0],seg_pre[4,15],
                                    seg_act[3,0],seg_pre[3,15],
                                    seg_act[2,0],
                                    seg_act[1,0],seg_pre[1,15],
                                    seg_act[0,1],seg_act[0,0],seg_pre[0,15],
                                    0]))
            hit_bin[15] =  list(map(int,[seg_suc[4,0],seg_act[4,15],seg_act[4,14],
                                    seg_act[3,15],seg_act[3,14],
                                    seg_act[2,15],
                                    seg_act[1,15],seg_act[1,14],
                                    seg_suc[0,0],seg_act[0,15],seg_act[0,14],
                                    0]))
        else:
            hit_bin[0] =  list(map(int,[seg_act[4,1],seg_act[4,0],0,
                                    seg_act[3,0],0,
                                    seg_act[2,0],
                                    seg_act[1,0],0,
                                    seg_act[0,1],seg_act[0,0],0,
                                    0]))
            hit_bin[15] =  list(map(int,[0,seg_act[4,15],seg_act[4,14],
                                    seg_act[3,15],seg_act[3,14],
                                    seg_act[2,15],
                                    seg_act[1,15],seg_act[1,14],
                                    0,seg_act[0,15],seg_act[0,14],
                                    0]))

        for i in range(1,16-1):
            hit_bin[i] =  list(map(int,[seg_act[4,i+1],seg_act[4,i],seg_act[4,i-1],
                                    seg_act[3,i],seg_act[3,i-1],
                                    seg_act[2,i],
                                    seg_act[1,i],seg_act[1,i-1],
                                    seg_act[0,i+1],seg_act[0,i],seg_act[0,i],
                                    0]))
        return hit_bin



    def _check_layer_hit(self,layer_in, edge_hit):
        """
        Checks frame layer <layer_in> for hits. Using edge_hit True/False segemnt edge mode can be
        choosen. For production use edge_hit=True as non edge_hit is only used for hardware module
        testing.
        """
        if(layer_in.layer_id == 0):
            n_segs = layer_in.n_segs
            hits = numpy.zeros(shape=(n_segs,16))
            # Segment 0
            seg_pre = layer_in.get_segment(n_segs-1)
            seg_act = layer_in.get_segment(0)
            seg_suc = layer_in.get_segment(1)
            hit_bin = self._populate_bin_inner(seg_pre, seg_act, seg_suc,edge_hit)
            hits[0] = numpy.array(list(map(self._detect_hit_inner, hit_bin)))
            # Inner segments
            for i in range (1,n_segs-1):
                seg_pre = layer_in.get_segment(i-1)
                seg_act = layer_in.get_segment(i)
                seg_suc = layer_in.get_segment(i+1)
                hit_bin = self._populate_bin_inner(seg_pre, seg_act, seg_suc,edge_hit)
                hits[i] = numpy.array(list(map(self._detect_hit_inner, hit_bin)))
            # Segment n_seg
            seg_pre = layer_in.get_segment(n_segs-2)
            seg_act = layer_in.get_segment(n_segs-1)
            seg_suc = layer_in.get_segment(0)
            hit_bin = self._populate_bin_inner(seg_pre, seg_act, seg_suc,edge_hit)
            hits[n_segs-1] = numpy.array(list(map(self._detect_hit_inner, hit_bin)))
        else:
            n_segs = layer_in.n_segs
            hits = numpy.zeros(shape=(n_segs,16))
            # Segment 0
            seg_pre = layer_in.get_segment(n_segs-1)
            seg_act = layer_in.get_segment(0)
            seg_suc = layer_in.get_segment(1)
            hit_bin = self._populate_bin_outer(seg_pre, seg_act, seg_suc,edge_hit)
            hits[0] = numpy.array(list(map(self._detect_hit_outer, hit_bin)))
            # Inner segments
            for i in range (1,n_segs-1):
                seg_pre = layer_in.get_segment(i-1)
                seg_act = layer_in.get_segment(i)
                seg_suc = layer_in.get_segment(i+1)
                hit_bin = self._populate_bin_outer(seg_pre, seg_act, seg_suc,edge_hit)
                hits[i] = numpy.array(list(map(self._detect_hit_outer, hit_bin)))
            # Segment n_seg
            seg_pre = layer_in.get_segment(n_segs-2)
            seg_act = layer_in.get_segment(n_segs-1)
            seg_suc = layer_in.get_segment(0)
            hit_bin = self._populate_bin_outer(seg_pre, seg_act, seg_suc,edge_hit)
            hits[n_segs-1] = numpy.array(list(map(self._detect_hit_outer, hit_bin)))

        return hits

    def _detect_hits(self,frame_in,edge_hit):
        """
        Detects hits of all layers in <frame_in> and saves output in tsf frame.
        Hardcoded to layer 0...8.

        Step 1: Calculate hits for each layer. Returns 2D Array of hits
                with indices <segment_id> whose range depends on the layer
                and 0...15 for each checked hit bin inside the dection algorithm.
                E.g. Bin 5 in Segment 3 was hit => hits[2,4] == True

        Step 2: Find all hits and fill output frame accordingly.
        """
        layer_id = 0
        layer_in = frame_in.get_layer(layer_id)
        layer_segments = self._frame.get_layer(layer_id).get_tsf_segments()
        hits = self._check_layer_hit(layer_in,edge_hit)

        for i in range(0,layer_in.n_segs):
            segment_id = i
            for j in range(0,16):
                if(hits[segment_id,j] == True):
                    layer_segments[segment_id,0,j] = 1

        for layer_id in range(1,9):
            layer_in = frame_in.get_layer(layer_id)
            layer_segments = self._frame.get_layer(layer_id).get_tsf_segments()
            hits = self._check_layer_hit(layer_in,edge_hit)

            for i in range(0,layer_in.n_segs):
                segment_id = i
                for j in range(0,16):
                    if(hits[segment_id,j] == True):
                        layer_segments[segment_id,2,j] = 1

        return self._frame


    def _detect_candidates(self,frame_in,edge_hit):
        """
        Detects hits of all layers in <frame_in> and saves output in tsf frame.
        Hardcoded to layer 1...8, because hardware implementation for layer 0
        is non existent. This feature could be implemented later.

        Step 1: Calculate hits for each layer. Returns 2D Array of hits
                with indices <segment_id> whose range depends on the layer
                and 0...15 for each checked hit bin inside the dection
                algorithm, e.g. Bin 5 in Segment 3 was hit => hits[2,4] == True

        Step 2: Find all hits and fill output frame accordingly.

        Identical to _detect_hit_outers() except that it populates all hit bin
        candidates to the internal output data frame.
        """
        layer_id = 0
        layer_in = frame_in.get_layer(layer_id)
        layer_segments = self._frame.get_layer(layer_id).get_tsf_segments()
        hits = self._check_layer_hit(layer_in,edge_hit)

        for i in range(0,layer_in.n_segs):
            segment_id = i
            for j in range(1,16-1):
                if(hits[segment_id,j] == True):
                    if(j==0):
                        layer_segments[segment_id-1,4,15] = 1
                        layer_segments[segment_id-1,4,14] = 1
                        layer_segments[segment_id,4,0] = 1
                        layer_segments[segment_id,4,1] = 1
                        layer_segments[segment_id,4,2] = 1
                        layer_segments[segment_id-1,3,15] = 1
                        layer_segments[segment_id,3,0] = 1
                        layer_segments[segment_id,3,1] = 1
                        layer_segments[segment_id,3,2] = 1
                        layer_segments[segment_id-1,2,15] = 1
                        layer_segments[segment_id,2,0] = 1
                        layer_segments[segment_id,2,1] = 1
                        layer_segments[segment_id,1,0] = 1
                        layer_segments[segment_id,1,1] = 1
                        layer_segments[segment_id,0,0] = 1
                    elif(j==1):
                        layer_segments[segment_id-1,4,15] = 1
                        layer_segments[segment_id,4,0] = 1
                        layer_segments[segment_id,4,1] = 1
                        layer_segments[segment_id,4,2] = 1
                        layer_segments[segment_id,4,3] = 1
                        layer_segments[segment_id,3,0] = 1
                        layer_segments[segment_id,3,1] = 1
                        layer_segments[segment_id,3,2] = 1
                        layer_segments[segment_id,3,3] = 1
                        layer_segments[segment_id,2,0] = 1
                        layer_segments[segment_id,2,1] = 1
                        layer_segments[segment_id,2,2] = 1
                        layer_segments[segment_id,1,1] = 1
                        layer_segments[segment_id,1,2] = 1
                        layer_segments[segment_id,0,1] = 1
                    elif(j==14):
                        layer_segments[segment_id,4,12] = 1
                        layer_segments[segment_id,4,13] = 1
                        layer_segments[segment_id,4,14] = 1
                        layer_segments[segment_id,4,15] = 1
                        layer_segments[segment_id+1 % layer_in.n_segs,4,0] = 1
                        layer_segments[segment_id,3,13] = 1
                        layer_segments[segment_id,3,14] = 1
                        layer_segments[segment_id,3,15] = 1
                        layer_segments[segment_id+1 % layer_in.n_segs,3,0] = 1
                        layer_segments[segment_id,2,13] = 1
                        layer_segments[segment_id,2,14] = 1
                        layer_segments[segment_id,2,15] = 1
                        layer_segments[segment_id,1,14] = 1
                        layer_segments[segment_id,1,15] = 1
                        layer_segments[segment_id,0,14] = 1
                    elif(j==15):
                        layer_segments[segment_id,4,13] = 1
                        layer_segments[segment_id,4,14] = 1
                        layer_segments[segment_id,4,15] = 1
                        layer_segments[segment_id+1 % layer_in.n_segs,4,0] = 1
                        layer_segments[segment_id+1 % layer_in.n_segs,4,1] = 1
                        layer_segments[segment_id,3,14] = 1
                        layer_segments[segment_id,3,15] = 1
                        layer_segments[segment_id+1 % layer_in.n_segs,3,0] = 1
                        layer_segments[segment_id+1 % layer_in.n_segs,3,1] = 1
                        layer_segments[segment_id,2,14] = 1
                        layer_segments[segment_id,2,15] = 1
                        layer_segments[segment_id+1 % layer_in.n_segs,2,0] = 1
                        layer_segments[segment_id,1,15] = 1
                        layer_segments[segment_id+1 % layer_in.n_segs,1,0] = 1
                        layer_segments[segment_id,0,15] = 1
                    else:
                        layer_segments[segment_id,4,j-2] = 1
                        layer_segments[segment_id,4,j-1] = 1
                        layer_segments[segment_id,4,j] = 1
                        layer_segments[segment_id,4,j+1] = 1
                        layer_segments[segment_id,4,j+2] = 1
                        layer_segments[segment_id,3,j-1] = 1
                        layer_segments[segment_id,3,j] = 1
                        layer_segments[segment_id,3,j+1] = 1
                        layer_segments[segment_id,3,j+2] = 1
                        layer_segments[segment_id,2,j-1] = 1
                        layer_segments[segment_id,2,j] = 1
                        layer_segments[segment_id,2,j+1] = 1
                        layer_segments[segment_id,1,j] = 1
                        layer_segments[segment_id,1,j+1] = 1
                        layer_segments[segment_id,0,j] = 1

        for layer_id in range(1,9):
            layer_in = frame_in.get_layer(layer_id)
            layer_segments = self._frame.get_layer(layer_id).get_tsf_segments()
            hits = self._check_layer_hit(layer_in,edge_hit)

            for i in range(0,layer_in.n_segs):
                segment_id = i
                for j in range(1,16-1):
                    if(hits[segment_id,j] == True):
                        if(j==0):
                            layer_segments[segment_id-1,4,15] = 1
                            layer_segments[segment_id,4,0] = 1
                            layer_segments[segment_id,4,1] = 1
                            layer_segments[segment_id-1,3,15] = 1
                            layer_segments[segment_id,3,0] = 1
                            layer_segments[segment_id,2,0] = 1
                            layer_segments[segment_id-1,1,15] = 1
                            layer_segments[segment_id,1,0] = 1
                            layer_segments[segment_id,0,1] = 1
                            layer_segments[segment_id-1,0,15] = 1
                            layer_segments[segment_id,0,0] = 1
                        elif(j==15):
                            layer_segments[segment_id,4,14] = 1
                            layer_segments[segment_id,4,15] = 1
                            layer_segments[segment_id+1 % layer_in.n_segs,4,0] = 1
                            layer_segments[segment_id,3,14] = 1
                            layer_segments[segment_id,3,15] = 1
                            layer_segments[segment_id,2,15] = 1
                            layer_segments[segment_id,1,14] = 1
                            layer_segments[segment_id,1,15] = 1
                            layer_segments[segment_id+1 % layer_in.n_segs,0,0] = 1
                            layer_segments[segment_id,0,14] = 1
                            layer_segments[segment_id,0,15] = 1
                        else:
                            layer_segments[segment_id,4,j-1] = 1
                            layer_segments[segment_id,4,j] = 1
                            layer_segments[segment_id,4,j+1] = 1
                            layer_segments[segment_id,3,j-1] = 1
                            layer_segments[segment_id,3,j] = 1
                            layer_segments[segment_id,2,j] = 1
                            layer_segments[segment_id,1,j-1] = 1
                            layer_segments[segment_id,1,j] = 1
                            layer_segments[segment_id,0,j+1] = 1
                            layer_segments[segment_id,0,j-1] = 1
                            layer_segments[segment_id,0,j] = 1
        return self._frame


    def get_frame(self):
        return self._frame


    def detect(self,frame_in,mode='hits',edge_hit=True):
        frame_out = None
        if(mode == 'hits'):
            self.mode = 'hits'
            frame_out = self._detect_hits(frame_in,edge_hit)
        elif(mode == 'candidates'):
            self._mode = 'candidates'
            frame_out = self._detect_candidates(frame_in,edge_hit)
        else:
            raise ValueError(mode)

        return frame_out


    def clear(self):
        self._frame = self._frame.clear()
        return self._frame



class LutConfig:
    """
    LutConfig creates a config class for LUT TrackSegmentFinder. It provides
    interfaces to save a configuration and to export it to a BRAM initialization
    file for Xilinx Ultrascale 7 FPGAs.
    """
    def __init__(self,size):
        if(size in range(16)):
            self._size = size
            self._n_bins = 2 ** size
            self._lutlist = [0 for i in range(self._n_bins)]
            self._patterns = []
        else:
            raise ValueError("size parameter to large or negativ. Choose a value in range(0,16).")

    def _list_from_file(self,path):
        """
        Helper function to read list object from file.
        """
        lines = []
        with open(path) as file_ptr:
            for line in file_ptr:
                line = line.strip()
                lines.append(line)
        return lines


    def _list_to_file(self,path,list_in):
        """
        Helper function to read list object from file.
        """
        with open(path, "w") as f:
            for line in list_in:
                f.write(f"{line}\n")


    def set_patterns(self,patterns):
        """
        Sets pattern object and updates internal data structure.
        """
        self._patterns = patterns
        for i in range(self._n_bins):
            self._lutlist[i] = f"{i:0{self._size:d}b}" in patterns


    def get_patterns(self):
        """
        Returns pattern object.
        """
        return self._patterns

    def add_pattern(self,pattern):
        """
        Adds a single pattern to the LutConfig Object.
        Parameter pattern: String with <size> number of chars, e.g.
        '1001' for size = 4.
        """
        self._patterns = self._patterns + [pattern]
        i = int(pattern, 2)
        self._lutlist[i] = 1


    def write_init(self,path):
        """
        Writes ram init configuration file for Xilinx Ultrascale 7 FPGAs.
        """
        file_path = f"{path}.init"
        with open(file_path,"w") as file_ptr:
            for data_line in self._lutlist:
                file_ptr.write(f"{data_line:1d}\n")


    def write_config(self,path):
        """
        Writes human readable config file under given <path>. Backwards
        compatible.
        """
        self._list_to_file(path,self._patterns)


    def read_config(self,path,n_patterns):
        """
        Reads human readable config file under given <path>. Backwards
        compatible. Read n patterns from file
        """
        new_patterns = self._list_from_file(path)[:n_patterns]
        if(self._n_bins >= n_patterns):
            self.set_patterns(new_patterns)
            return self
        else:
            raise ValueError("Config File length does not match LUTConfig size.")



import pkg_resources
import numpy
from pandas import read_csv
from plotly.io import write_image
import plotly.graph_objects as go
from plotly.graph_objs import Layout
from b2tsf.belle_CDC_frame import CDCFrame


class CDCCanvas:
    """
    Class containing a plotly canvas for visualization of data and track segment
    finder algorithm results. Handles all the low level things for drawing nice
    pictures. When setting up specify:
        - path to the coordinate lookup table (wire_dict_path).

        - to draw the inner layer (skip_inner_layer).

        - whether to scale markers for a more clear visualization at a cost of
          physical accuracy (scale_wires).
    ----------------------------------------------------------------------------
    Methods:
    ----------------------------------------------------------------------------
    draw:   Draws CDCFrame data onto the canvas in a mode specified. The
            following modes are available (mode=):

            -'normal' Draws hit and no hit data is predefined marker and color
                      schemes. Use this to draw input data.

            -'tsf'    Draw tsf hit data. Best pratice is to draw the data in
                      'normal' mode first and then include tsf hit data on top.

    plot:   Plot canvas to file specified in path. Caution: Files might be large
            depending on datapoints. The following modes are available(mode=):

            -'all'
                plot the whole 360° frame

            -'top','bottom','left','right'
                plot 180° of a frame

            -'top/bottom'-'left/right'
                plot 90° of a frame

    clear:  Clears the canvas from all data.

    """

    def __init__(self,
                 wire_dict_path='data/tsf_coordinates_translation.txt',
                 skip_inner_layer=False,
                 scale_wires=True):
        self._no_hit_color='#4664AA'
        self._wire_hit_color='#009682'
        self._tsf_hit_color='#A3107C'
        self._skip_inner_layer=skip_inner_layer
        self._scale_wires=scale_wires
        self._figure = go.Figure()
        self._frame = CDCFrame(0)

        file_path = pkg_resources.resource_filename(__name__, 'data/tsf_coordinates_translation.txt')
        co_data = read_csv(file_path,delimiter=',')
        keys = []
        values_x = []
        values_y = []
        for index, row in co_data.iterrows():
            new_key = (row['superlayer id'],row['ilayer id'],row['azimuthal id'])
            keys.append(new_key)
            values_x.append(row['x'])
            values_y.append(row['y'])

        self._x_dict = dict(zip(keys,values_x))
        self._y_dict = dict(zip(keys,values_y))


    def _retrieve_x_coordinates(self,tupel_id):
        return self._x_dict[tupel_id]

    def _retrieve_y_coordinates(self,tupel_id):
        return self._y_dict[tupel_id]

    def _draw_normal(self):
        hits = list(zip(*self._frame.get_hit_list(format='r_theta',
            detect_hit=True, skip_inner_layer=self._skip_inner_layer)))
        x = list(map(self._retrieve_x_coordinates,hits))
        y = list(map(self._retrieve_y_coordinates,hits))
        if(self._scale_wires is True):
            r = numpy.sqrt(numpy.array(x)**2+numpy.array(y)**2)
        else:
            r = 0

        self._figure.add_trace(go.Scatter(
            name = 'Wire hit',
            legendgroup = 'Wire hit',
            x = x,
            y = y,
            mode = 'markers',
            marker_symbol='circle',
            marker_size=6 + r/180,
            marker_color=self._wire_hit_color,
            marker_line_color=self._wire_hit_color,
            showlegend = False,
        ))

        hits = list(zip(*self._frame.get_hit_list(format='r_theta',
            detect_hit=False, skip_inner_layer=self._skip_inner_layer)))
        x = list(map(self._retrieve_x_coordinates,hits))
        y = list(map(self._retrieve_y_coordinates,hits))
        if(self._scale_wires is True):
            r = numpy.sqrt(numpy.array(x)**2+numpy.array(y)**2)
        else:
            r = 0

        self._figure.add_trace(go.Scatter(
             name = 'No hit',
             legendgroup =  'No hit',
             x = x,
             y = y,
             mode = 'markers',
             marker_symbol='circle-open',
             marker_size=6 + r/180,
             marker_color=self._no_hit_color,
             showlegend=False,
        ))

    def _draw_tsf(self):
        hits = list(zip(*self._frame.get_hit_list(format='r_theta',
            detect_hit=True, skip_inner_layer=self._skip_inner_layer)))
        x = list(map(self._retrieve_x_coordinates,hits))
        y = list(map(self._retrieve_y_coordinates,hits))
        if(self._scale_wires is True):
            r = numpy.sqrt(numpy.array(x)**2+numpy.array(y)**2)
        else:
            r = 0

        self._figure.add_trace(go.Scatter(
             name = 'TSF hit',
             legendgroup =  'TSF hit',
             x = x,
             y = y,
             mode = 'markers',
             marker_symbol='circle-x-open',
             marker_size=6 + r/180,
             marker_color=self._tsf_hit_color,
             showlegend=False,
        ))


    def draw(self,frame,hit_type='normal'):
        self._frame = frame
        if(hit_type == 'normal'):
            self._draw_normal()
            return self
        elif(hit_type == 'tsf'):
            self._draw_tsf()
            return self
        else:
            return ValueError(hit_type)


    def plot(self,path='../out/new_plot.pdf',mode='all'):
        if(mode == 'all'):
            x_range = [-1200,1200]
            y_range = [-1200,1200]
            x_width = 2500
            y_width = 2500
        elif(mode == 'top-right'):
            x_range = [0,1200]
            y_range = [0,1200]
            x_width = 1250
            y_width = 1250
        elif(mode == 'top-left'):
            x_range = [-1200,0]
            y_range = [0,1200]
            x_width = 1250
            y_width = 1250
        elif(mode == 'bottom-right'):
            x_range = [0,1200]
            y_range = [0,-1200]
            x_width = 1250
            y_width = 1250
        elif(mode == 'bottom-left'):
            x_range = [-1200,0]
            y_range = [-1200,0]
            x_width = 1250
            y_width = 1250
        elif(mode == 'top'):
            x_range = [-1200,1200]
            y_range = [0,1200]
            x_width = 2500
            y_width = 1250
        elif(mode == 'bottom'):
            x_range = [-1200,0]
            y_range = [-1200,1200]
            x_width = 2500
            y_width = 1250
        elif(mode == 'right'):
            x_range = [0,1200]
            y_range = [-1200,1200]
            x_width = 1250
            y_width = 2500
        elif(mode == 'left'):
            x_range = [-1200,0]
            y_range = [-1200,1200]
            x_width = 1250
            y_width = 2500
        else:
            return ValueError(mode)

        self._figure.update_xaxes(
            dtick = 100,
            range = x_range,
            tickfont_size= 20,
            title_text = "x in mm",
            title_font = {"size": 32},
            title_standoff = 25)

        self._figure.update_yaxes(
            dtick = 100,
            range = y_range,
            tickfont_size = 20,
            title_text = "y in mm",
            title_font = {"size": 32},
            title_standoff = 25)

        layout = Layout(
            paper_bgcolor='rgba(0,0,0,0)',
            plot_bgcolor='rgba(70,100,170,0.1)',

        )
        self._figure.update_layout(layout)
        write_image(self._figure,path,width=x_width,height=y_width)

    def clear(self):
        self._frame = CDCFrame(0)
        self._figure = go.Figure()
